import os
import sys
import unittest
from pypdn.nrbf import NRBF

# TODO Write NRBF unit tests
class TestNRBF(unittest.TestCase):
    def setUp(self):
        pass

    def test_read_nrbf(self):
        pass


if __name__ == '__main__':
    unittest.main()



import os
import sys
import unittest

import imageio

from pypdn.reader import *
from tests.util import getDataPath


class TestReader(unittest.TestCase):
    def setUp(self):
        self.layerVisibleTest1 = imageio.imread(getDataPath('layerVisibleTest1.png'), ignoregamma=True)
        self.layerVisibleTest2 = imageio.imread(getDataPath('layerVisibleTest2.png'), ignoregamma=True)
        self.layerVisibleTest3 = imageio.imread(getDataPath('layerVisibleTest3.png'), ignoregamma=True)
        self.flattenNormalTest = imageio.imread(getDataPath('flattenNormalTest.png'), ignoregamma=True)
        self.flattenNormalTest2 = imageio.imread(getDataPath('flattenNormalTest2.png'), ignoregamma=True)

        self.flattenMultiplyTest = imageio.imread(getDataPath('flattenMultiplyTest.png'), ignoregamma=True)
        self.flattenAdditiveTest = imageio.imread(getDataPath('flattenAdditiveTest.png'), ignoregamma=True)
        self.flattenColorBurnTest = imageio.imread(getDataPath('flattenColorBurnTest.png'), ignoregamma=True)
        self.flattenColorDodgeTest = imageio.imread(getDataPath('flattenColorDodgeTest.png'), ignoregamma=True)
        self.flattenReflectTest = imageio.imread(getDataPath('flattenReflectTest.png'), ignoregamma=True)
        self.flattenGlowTest = imageio.imread(getDataPath('flattenGlowTest.png'), ignoregamma=True)
        self.flattenOverlayTest = imageio.imread(getDataPath('flattenOverlayTest.png'), ignoregamma=True)
        self.flattenDifferenceTest = imageio.imread(getDataPath('flattenDifferenceTest.png'), ignoregamma=True)
        self.flattenNegationTest = imageio.imread(getDataPath('flattenNegationTest.png'), ignoregamma=True)
        self.flattenLightenTest = imageio.imread(getDataPath('flattenLightenTest.png'), ignoregamma=True)
        self.flattenDarkenTest = imageio.imread(getDataPath('flattenDarkenTest.png'), ignoregamma=True)
        self.flattenScreenTest = imageio.imread(getDataPath('flattenScreenTest.png'), ignoregamma=True)
        self.flattenXORTest = imageio.imread(getDataPath('flattenXORTest.png'), ignoregamma=True)

        self.flattenOpacityTest = imageio.imread(getDataPath('flattenOpacityTest.png'), ignoregamma=True)

    def test_read(self):
        layeredImage = read(getDataPath('Untitled3.pdn'))

        self.assertEqual(layeredImage.width, 800)
        self.assertEqual(layeredImage.height, 600)

        self.assertEqual(layeredImage.version.__class__.__name__, 'System_Version')
        self.assertTrue(hasattr(layeredImage.version, 'Major'))
        self.assertTrue(hasattr(layeredImage.version, 'Minor'))
        self.assertTrue(hasattr(layeredImage.version, 'Build'))
        self.assertTrue(hasattr(layeredImage.version, 'Revision'))
        self.assertIsInstance(layeredImage.version.Major, int)
        self.assertIsInstance(layeredImage.version.Minor, int)
        self.assertIsInstance(layeredImage.version.Build, int)
        self.assertIsInstance(layeredImage.version.Revision, int)

        self.assertEqual(len(layeredImage.layers), 2)

        layer = layeredImage.layers[0]
        self.assertEqual(layer.name, 'Background')
        self.assertEqual(layer.visible, True)
        self.assertEqual(layer.isBackground, True)
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)
        np.testing.assert_equal(layer.image, self.layerVisibleTest1)

        layer = layeredImage.layers[1]
        self.assertEqual(layer.name, 'Layer 2')
        self.assertEqual(layer.visible, True)
        self.assertEqual(layer.isBackground, False)
        self.assertEqual(layer.opacity, 161)
        self.assertEqual(layer.blendMode, BlendType.Additive)
        # Ignore the alpha component because it should be 255 and not 161 like in image that is flattened
        np.testing.assert_equal(layer.image[:, :, 0:3], self.layerVisibleTest2[:, :, 0:3])

    def test_flatten_normal1(self):
        layeredImage = read(getDataPath('Untitled2.pdn'))

        self.assertEqual(len(layeredImage.layers), 2)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        layer = layeredImage.layers[1]
        layer.visible = False
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        image = layeredImage.flatten(asByte=True)
        np.testing.assert_equal(image, self.flattenNormalTest2)

    def test_flatten_normal2(self):
        layeredImage = read(getDataPath('Untitled2.pdn'))

        self.assertEqual(len(layeredImage.layers), 2)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        layer = layeredImage.layers[1]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        image = layeredImage.flatten(asByte=True)
        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenNormalTest, atol=2)

    def test_flatten_notasbyte(self):
        layeredImage = read(getDataPath('Untitled2.pdn'))

        self.assertEqual(len(layeredImage.layers), 2)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        layer = layeredImage.layers[1]
        layer.visible = False
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        image = layeredImage.flatten(asByte=False)
        np.testing.assert_equal(image, self.flattenNormalTest2 / 255.)

    def test_flatten_multiply(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Multiply
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # plt.imshow(image)
        # plt.show()
        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenMultiplyTest, atol=2)

    def test_flatten_additive(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Additive
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenAdditiveTest, atol=2)

    def test_flatten_colorBurn(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.ColorBurn
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenColorBurnTest, atol=2)

    def test_flatten_colorDodge(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.ColorDodge
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenColorDodgeTest, atol=2)

    def test_flatten_reflect(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Reflect
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenReflectTest, atol=2)

    def test_flatten_glow(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Glow
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenGlowTest, atol=2)

    def test_flatten_overlay(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Overlay
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenOverlayTest, atol=2)

    def test_flatten_difference(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Difference
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenDifferenceTest, atol=2)

    def test_flatten_negation(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Negation
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenNegationTest, atol=2)

    def test_flatten_lighten(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Lighten
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenLightenTest, atol=2)

    def test_flatten_darken(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Darken
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenDarkenTest, atol=2)

    def test_flatten_screen(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.Screen
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenScreenTest, atol=2)

    def test_flatten_xor(self):
        layeredImage = read(getDataPath('FlattenBlendTest.pdn'))

        self.assertEqual(len(layeredImage.layers), 14)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        desiredLayer = BlendType.XOR
        for x in range(1, 14):
            layeredImage.layers[x].visible = (x == desiredLayer)

        self.assertEqual(layeredImage.layers[desiredLayer].opacity, 255)
        self.assertEqual(layeredImage.layers[desiredLayer].blendMode, desiredLayer)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenXORTest, atol=2)

    def test_flatten_opacity(self):
        layeredImage = read(getDataPath('Untitled3.pdn'))

        self.assertEqual(len(layeredImage.layers), 2)

        layer = layeredImage.layers[0]
        layer.visible = True
        self.assertEqual(layer.opacity, 255)
        self.assertEqual(layer.blendMode, BlendType.Normal)

        layer = layeredImage.layers[1]
        layer.visible = True
        self.assertEqual(layer.opacity, 161)
        self.assertEqual(layer.blendMode, BlendType.Additive)

        image = layeredImage.flatten(asByte=True)

        # May be rounding errors so do within 2 points
        np.testing.assert_allclose(image, self.flattenOpacityTest, atol=2)


if __name__ == '__main__':
    unittest.main()



import gzip
import struct

import numpy as np
import skimage
from aenum import IntEnum
import warnings

from pypdn.nrbf import NRBF


# TODO Add API for writing PDN file

# TODO Write docstrings basic

class PDNReaderError(Exception):
    """Exceptions for Paint.NET Reader module."""
    pass


class BlendType(IntEnum):
    Normal = 0
    Multiply = 1
    Additive = 2
    ColorBurn = 3
    ColorDodge = 4
    Reflect = 5
    Glow = 6
    Overlay = 7
    Difference = 8
    Negation = 9
    Lighten = 10
    Darken = 11
    Screen = 12
    XOR = 13


class LayeredImage:
    __slots__ = ['width', 'height', 'version', 'layers']

    def __init__(self, width, height, version):
        self.width = width
        self.height = height
        self.version = version
        self.layers = []

    def __repr__(self):
        return 'pypdn.LayeredImage(width={0}, height={1}, version={2!r}, layers={3!r})'.format(self.width, self.height,
                                                                                             self.version, self.layers)

    def flatten(self, useAll=False, applyBlendMode=True, asByte=False):
        if len(self.layers) == 0:
            return None

        # Create empty RGBA image to hold the flattened image
        # Use float datatype because that is how the calculations are done to prevent scaling
        # Image can be scaled at the end
        image = np.zeros((self.height, self.width, 4), dtype=float)

        for layer in self.layers:
            if useAll or layer.visible:
                blendMode = BlendType.Normal if not applyBlendMode else layer.blendMode

                # Operations must be done on float image!
                normalizedImage = skimage.img_as_float(layer.image)

                # Apply the layer opacity to the image here
                # If the image does not have an alpha component, extend the image to contain one
                if normalizedImage.shape[2] == 3:
                    alpha = np.ones(normalizedImage.shape[:-1], dtype=float)
                    normalizedImage = np.dstack((normalizedImage, alpha))

                # Now multiply the layer opacity by the alpha component of image
                normalizedImage[:, :, 3] = normalizedImage[:, :, 3] * (layer.opacity / 255.)

                # Take current image and apply the normalized image from the layer to it with specified blend mode
                image = applyBlending(image, normalizedImage, blendMode)

        # Paint.NET stores everything as uint8's
        # We had to convert to float to flatten the image and now we can convert back to uint8 if desired
        if asByte:
            # Catch warnings to prevent the possible precision lost warning
            # Personally, I think the user should know that converting to a uint8 will cause precision lost from float
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                image = skimage.img_as_ubyte(image)

        return image


class Layer:
    __slots__ = ['name', 'visible', 'isBackground', 'opacity', 'blendMode', 'image']

    def __init__(self, name, visible, isBackground, opacity, blendMode, image):
        self.name = name
        self.visible = visible
        self.isBackground = isBackground
        self.opacity = opacity
        self.blendMode = blendMode
        self.image = image

    def __repr__(self):
        return 'pypdn.Layer(name={0}, visible={1}, isBackground={2}, opacity={3}, blendMode={4!r})' \
            .format(self.name, self.visible, self.isBackground, self.opacity, self.blendMode)


def read(filename):
    with open(filename, 'rb') as fh:
        # Begin by reading magic str and checking it
        magicStr = fh.read(4).decode('ascii')

        if magicStr != 'PDN3':
            raise PDNReaderError('Invalid magic string for PDN file: %s' % magicStr)

        headerSizeStr = fh.read(3) + b'\x00'
        if len(headerSizeStr) != 4:
            raise PDNReaderError('Unable to read header size. File may be corrupted.')

        # Read header size and the header
        headerSize = struct.unpack('<i', headerSizeStr)[0]
        header = fh.read(headerSize).decode('utf-8')

        # Note: The header does not contain any relevant information that is not contained in the NRBF part itself
        # Only section is the thumbnail image but that is not relevant here I do not believe

        if fh.read(2) != b'\x00\x01':
            raise PDNReaderError('Invalid data indicator bytes. File may be corrupted')

        nrbfData = NRBF(stream=fh)
        pdnDocument = nrbfData.getRoot()

        try:
            layeredImage = LayeredImage(pdnDocument.width, pdnDocument.height, pdnDocument.savedWith)

            # Cannot loop through items array because sometimes it is padded with null objects, size is right though
            for index in range(pdnDocument.layers.ArrayList__size):
                bitmapLayer = pdnDocument.layers.ArrayList__items[index]
                layerProps = bitmapLayer.Layer_properties

                # Read information from layer that is used to read the image data
                assert bitmapLayer.Layer_width == layeredImage.width
                assert bitmapLayer.Layer_height == layeredImage.height
                stride = bitmapLayer.surface.stride
                length = bitmapLayer.surface.scan0.length64

                # Begin reading the image data from the layer
                # Look here for a reference of how it is written:
                # https://github.com/rivy/OpenPDN/blob/cca476b0df2a2f70996e6b9486ec45327631568c/src/Core/MemoryBlock.cs

                # Empty array of the length, will be filled in later
                data = bytearray([0] * length)

                # Format version of the data, 0 = GZIP compressed, 1 = uncompressed
                formatVersion = struct.unpack('>B', fh.read(1))[0]
                # Size of each chunk in the destination (buffer where we store data)
                chunkSize = struct.unpack('>I', fh.read(4))[0]

                # Get total number of chunks which is total length divided by chunkSize
                # Keep track of the chunks found in case the file is corrupted
                chunkCount = np.ceil(length / chunkSize).astype(np.uint32)
                chunksFound = [False] * chunkCount

                for x in range(chunkCount):
                    # Read the chunk number, they are not necessarily in order
                    chunkNumber = struct.unpack('>I', fh.read(4))[0]

                    if chunkNumber >= chunkCount:
                        raise PDNReaderError(
                            'Chunk number read from stream is out of bounds: %i %i. File may be corrupted'
                            % (chunkNumber, chunkCount))

                    if chunksFound[chunkNumber]:
                        raise PDNReaderError(
                            'Chunk number %i was already encountered. File may be corrupted' % chunkNumber)

                    # Read the size of the data from memory
                    # This is not necessary the same size as chunkSize if the data is compressed
                    dataSize = struct.unpack('>I', fh.read(4))[0]

                    # Calculate the chunk offset
                    chunkOffset = chunkNumber * chunkSize

                    # Mark this chunk as found
                    chunksFound[chunkNumber] = True

                    # The chunk size is a maximum value and should be limited for the last chunk where it might not be
                    # exactly equal to the chunk size
                    actualChunkSize = np.min((chunkSize, length - chunkOffset))

                    # Read the chunk data
                    rawData = fh.read(dataSize)

                    if formatVersion == 0:
                        decompressedData = gzip.decompress(rawData)
                        data[chunkOffset:chunkOffset + actualChunkSize] = decompressedData
                    else:
                        data[chunkOffset:chunkOffset + actualChunkSize] = rawData

                # With the data read in as one large 1D array, we now format the data properly
                # Calculate the bits per pixel
                bpp = stride * 8 / layeredImage.width

                # Convert 1D array to image array
                if bpp == 32:
                    image = np.frombuffer(data, np.uint8).reshape((layeredImage.height, layeredImage.width, 4))
                    image[:, :, 0:3] = np.flip(image[:, :, 0:3], axis=-1)
                elif bpp == 24:
                    image = np.frombuffer(data, np.uint8).reshape((layeredImage.height, layeredImage.width, 3))
                    image[:, :, 0:2] = np.flip(image[:, :, 0:2], axis=-1)
                else:
                    raise PDNReaderError('Invalid bpp %i' % bpp)

                # Setup layer with information and append to the layers list
                layer = Layer(layerProps.name, layerProps.visible, layerProps.isBackground, layerProps.opacity,
                              BlendType(layerProps.blendMode.value__), image)
                layeredImage.layers.append(layer)

            return layeredImage
        except (AttributeError):
            raise PDNReaderError('Unable to read fields in NRBF PDN file')


def blendingFunc(A, B, blendType):
    # See here for information on Paint.NET's blending functions:
    # https://github.com/rivy/OpenPDN/blob/master/src/Data/UserBlendOps.Generated.H.cs
    if blendType == BlendType.Normal:
        return B
    elif blendType == BlendType.Multiply:
        return A * B
    elif blendType == BlendType.Additive:
        return np.minimum(A + B, 1.0)
    elif blendType == BlendType.ColorBurn:
        with np.errstate(divide='ignore'):
            return np.where(B != 0.0, np.maximum(1.0 - ((1.0 - A) / B), 0.0), 0.0)
    elif blendType == BlendType.ColorDodge:
        with np.errstate(divide='ignore'):
            return np.where(B != 1.0, np.minimum(A / (1.0 - B), 1.0), 1.0)
    elif blendType == BlendType.Reflect:
        with np.errstate(divide='ignore'):
            return np.where(B != 1.0, np.minimum((A ** 2) / (1.0 - B), 1.0), 1.0)
    elif blendType == BlendType.Glow:
        with np.errstate(divide='ignore'):
            return np.where(A != 1.0, np.minimum((B ** 2) / (1.0 - A), 1.0), 1.0)
    elif blendType == BlendType.Overlay:
        return np.where(A < 0.5, 2 * A * B, 1.0 - (2 * (1.0 - A) * (1.0 - B)))
    elif blendType == BlendType.Difference:
        return np.abs(A - B)
    elif blendType == BlendType.Negation:
        return 1.0 - np.abs(1.0 - A - B)
    elif blendType == BlendType.Lighten:
        return np.maximum(A, B)
    elif blendType == BlendType.Darken:
        return np.minimum(A, B)
    elif blendType == BlendType.Screen:
        return A + B - A * B
    elif blendType == BlendType.XOR:
        # XOR is meant for integer numbers, so must convert to uint8 first
        # Catch warnings to prevent the precision lost warning
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return skimage.img_as_float(skimage.img_as_ubyte(A) ^ skimage.img_as_ubyte(B))


def applyBlending(A, B, blendType):
    # The generalized alpha composite with a changeable blending function is shown below
    #
    # G(A,a,B,b,F) = (a - ab)A + (b - ab)B + abF(A, B)
    #
    # Where:
    #   A = background color
    #   a = background alpha
    #   B = foreground color
    #   b = foreground alpha
    #   F is blending function (F(A,B))
    #
    # This is what Paint.NET uses to create the flattened image and I do so as well but with Numpy to make it
    # easier.
    aAlpha = A[:, :, 3]
    bAlpha = B[:, :, 3]
    abAlpha = aAlpha * bAlpha

    aColor = A[:, :, 0:3]
    bColor = B[:, :, 0:3]

    # Use generalized alpha compositing equation shown above to calculate the color components of the image
    # To be able to multiply along each component (R,G,B), we must add a new axis to the difference portion
    # That is what the None field indicates
    colorComponents = (aAlpha - abAlpha)[:, :, None] * aColor + \
                      (bAlpha - abAlpha)[:, :, None] * bColor + \
                      abAlpha[:, :, None] * blendingFunc(aColor, bColor, blendType)

    # #define COMPUTE_ALPHA(a, b, r) { INT_SCALE(a, 255 - (b), r); r += (b); }
    # alphaComponent = aAlpha * (1.0 - bAlpha) + bAlpha
    alphaComponent = aAlpha + bAlpha - abAlpha

    # Return RGBA image by combining RGB and A
    return np.dstack((colorComponents, alphaComponent))



import struct
from datetime import timezone
from functools import reduce

from pypdn.util import *


# This library was inspired from the following code:
#   https://github.com/gurnec/Undo_FFG/blob/master/nrbf.py
# Thanks to Christopher Gurnee!

# TODO Do documentation and docstrings for this
# TODO Issue with reporting _id in JSON or something. classID vs _id is issue

class NRBFError(Exception):
    """Exceptions for NRBFError class."""
    pass


# Decorator which adds an enum value (an int or a length-one bytes object) and its associated reader function to a
# registration dict (either PrimitiveTypeReaders or RecordTypeReaders)
# The last two arguments, primitiveStructs and primitiveFormat are only specified for primitive types to create the
# Struct object. The Struct object parses the format and makes it quicker to parse the numbers by calling Struct.unpack
def _registerReader(typeDict, typeValue, primitiveStructs=None, primitiveFormat=None):
    assert isinstance(typeValue, int)

    def decorator(readFunction):
        if primitiveStructs is not None:
            primitiveStructs[typeValue] = struct.Struct(primitiveFormat)

        typeDict[typeValue] = readFunction
        return readFunction

    return decorator


class NRBF:
    # Dictionary that contains functions to call when reading records and primitives from NRBF file
    _RecordTypeReaders = {}
    _PrimitiveTypeReaders = {}

    # Dictionary that contains Struct objects for each of the primitive types. The Struct objects are precompiled to
    # speed up parsing the number
    _PrimitiveTypeStructs = {}

    # Dictionary contains stuff for reading primitive arrays
    _PrimitiveTypeArrayReaders = {}

    # Dictionary containing readers for additional info in classes
    _AdditionalInfoReaders = {}

    def __init__(self, stream=None, filename=None):
        self.stream = None
        self.rootID = None
        self.headerID = None
        self.referencesResolved = None

        # Dictionary of binary libraries with their linked objects
        self.binaryLibraries = {}

        # Keep track of class and objects by their ID
        self.classByID = {}
        self.objectsByID = {}

        # Keeps track of references so that after reading is done, the references can be resolved
        # collection references replace the system collections with Python equivalent types such as dict or list
        self._collectionReferences = []

        # If a stream or filename to be loaded is given, then call the read function
        # This makes the syntax cleaner to allow this creation of class and loading of data all in one line
        if stream is not None or filename is not None:
            self.read(stream, filename)

    def read(self, stream=None, filename=None):
        if stream is None and filename is not None:
            stream = open(filename, 'rb')

        assert stream.readable()

        if self.stream is not None:
            # Means we are in the middle of reading or writing data, throw exception probably
            raise NRBFError('Class is already reading from a stream! Please close the stream before trying again')

        if self.rootID is not None:
            # File has already been loaded so we must reset everything
            self.rootID = None
            self.headerID = None
            self.binaryLibraries = {}
            self.classByID = {}
            self.objectsByID = {}
            self._collectionReferences = []

        self.stream = stream

        # Read header
        self._readRecord()
        if self.rootID is None:
            raise NRBFError('Invalid stream, unable to read header. File may be corrupted')

        # Keep reading records until we receive a MessageEnd record
        while not isinstance(self._readRecord(), MessageEnd):
            pass

        # Resolve all the collection references
        self.resolveReferences()

        # Once we are done reading, we set the stream to None because it will not be used
        self.stream = None

    def write(self, stream=None, filename=None):
        if stream is None and filename is not None:
            stream = open(filename, 'wb')

        assert stream.writable()
        assert stream.seekable()

        # TODO Write NRBF files
        raise NotImplementedError('Writing a NRBF file is not supported yet')

    def getRoot(self):
        assert self.rootID is not None

        return self.objectsByID[self.rootID]

    def resolveReferences(self):
        # Resolve all the collection references
        for reference in self._collectionReferences:
            # Calls one of the resolvers
            replacement = reference.collectionResolver(self, reference)

            # The final steps common to all collection resolvers are completed below
            if reference.parent:
                reference.parent[reference.index_in_parent] = replacement

            self.objectsByID[reference._id] = replacement

        # Note: Collection references must be saved so that it can be converted back
        # when saving the file again.

        # Loop through all of the objects
        for _, object in self.objectsByID.items():
            # Attempt to iterate through the object
            # If it doesnt work, then move on to the next item because there wont need to be any
            # references.
            # Otherwise, if any of the items have an object ID, create a reference
            try:
                for index, item in enumerate(object):
                    if isinstance(item, Reference):
                        self._resolveSimpleReference(item)
            except TypeError:
                pass

        self.referencesResolved = True

    def unresolveReferences(self):
        # Loop through all of the objects
        for _, object in self.objectsByID.items():
            # Attempt to iterate through the object
            # If it doesnt work, then move on to the next item because there wont need to be any
            # references.
            # Otherwise, if any of the items have an object ID, create a reference
            try:
                for index, item in enumerate(object):
                    if hasattr(item, '_id'):
                        object[index] = Reference(item._id, object, index)
            except TypeError:
                pass

        self.referencesResolved = False

    def toJSON(self, resolveReferences=True, **kwargs):
        # Resolve or unresolve the references based on what the user desires
        # and the current state
        if resolveReferences and not self.referencesResolved:
            self.resolveReferences()
        elif not resolveReferences and self.referencesResolved:
            self.unresolveReferences()

        jsonEncoder = JSONEncoder(**kwargs)
        return jsonEncoder.encode(self)

    # region Primitive reader functions

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Boolean, _PrimitiveTypeStructs, '?')
    def _readBool(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Boolean].unpack(self.stream.read(1))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Byte, _PrimitiveTypeStructs, 'B')
    @_registerReader(_AdditionalInfoReaders, BinaryType.Primitive)
    @_registerReader(_AdditionalInfoReaders, BinaryType.PrimitiveArray)
    def _readByte(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Byte].unpack(self.stream.read(1))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Char)
    def _readChar(self):
        utf8Bytes = bytearray()
        while True:
            utf8Bytes += self.stream.read(1)
            try:
                return utf8Bytes.decode('utf-8')
            except UnicodeDecodeError:
                if len(utf8Bytes) > 4:
                    raise NRBFError('Invalid char read from NRBF file, longer than 4 bytes: {0}'.format(utf8Bytes))

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Decimal)
    def _read_Decimal(self):
        return Decimal(self._readString())

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Double, _PrimitiveTypeStructs, '<d')
    def _readDouble(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Double].unpack(self.stream.read(8))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Int16, _PrimitiveTypeStructs, '<h')
    def _readInt16(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Int16].unpack(self.stream.read(2))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Int32, _PrimitiveTypeStructs, '<i')
    def _readInt32(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Int32].unpack(self.stream.read(4))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Int64, _PrimitiveTypeStructs, '<q')
    def _readInt64(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Int64].unpack(self.stream.read(8))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.SByte, _PrimitiveTypeStructs, 'b')
    def _readSByte(self):
        return self._PrimitiveTypeStructs[PrimitiveType.SByte].unpack(self.stream.read(1))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Single, _PrimitiveTypeStructs, '<f')
    def _readSingle(self):
        return self._PrimitiveTypeStructs[PrimitiveType.Single].unpack(self.stream.read(4))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.TimeSpan)
    def _readTimeSpan(self):
        # 64-bit integer that represents time span in increments of 100 nanoseconds
        # Divide by 10 to get into microseconds
        return timedelta(microseconds=self._readInt64() / 10)

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.DateTime)
    def _readDateTime(self):
        ticks = self._readUInt64()

        # Two MSB store kind, 0 = no timezone, 1 = UTC, 2 = local timezone
        kind = ticks >> 62
        # Remaining 62-bits are the number of 100ns increments from 12:00:00 January 1, 0001
        ticks &= (1 << 62) - 1
        # If negative, then reinterpret as 62-bit two's complement
        if ticks >= 1 << 61:
            ticks -= 1 << 62

        # Create a datetime that starts at the beginning and then increment it by the number of microseconds
        time = datetime(1, 1, 1)
        try:
            time += timedelta(microseconds=ticks / 10)
        except OverflowError:
            pass

        # Update datetime object to have the appropriate timezone
        # If kind is 1, then this is UTC and if kind is 2, then this is local timezone
        if kind == 1:
            time = time.replace(tzinfo=timezone.utc)
        elif kind == 2:
            LOCAL_TIMEZONE = datetime.now(timezone.utc).astimezone().tzinfo
            time = time.replace(tzinfo=LOCAL_TIMEZONE)  # kind 2 is the local time zone

        return time

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.UInt16, _PrimitiveTypeStructs, '<H')
    def _readUInt16(self):
        return self._PrimitiveTypeStructs[PrimitiveType.UInt16].unpack(self.stream.read(2))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.UInt32, _PrimitiveTypeStructs, '<I')
    def _readUInt32(self):
        return self._PrimitiveTypeStructs[PrimitiveType.UInt32].unpack(self.stream.read(4))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.UInt64, _PrimitiveTypeStructs, '<Q')
    def _readUInt64(self):
        return self._PrimitiveTypeStructs[PrimitiveType.UInt64].unpack(self.stream.read(8))[0]

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.Null)
    @_registerReader(_AdditionalInfoReaders, BinaryType.String)
    @_registerReader(_AdditionalInfoReaders, BinaryType.Object)
    @_registerReader(_AdditionalInfoReaders, BinaryType.ObjectArray)
    @_registerReader(_AdditionalInfoReaders, BinaryType.StringArray)
    @_registerReader(_RecordTypeReaders, RecordType.ObjectNull)
    def _readNull(self):
        return None

    @_registerReader(_PrimitiveTypeReaders, PrimitiveType.String)
    @_registerReader(_AdditionalInfoReaders, BinaryType.SystemClass)
    def _readString(self):
        length = 0

        # Each bit range is 7 bits long with a maximum of 5 bytes
        for bit_range in range(0, 5 * 7, 7):
            byte = self._readByte()

            # Remove the last bit from the length (used to indicate if there is another byte to come)
            # Then shift the number to the appropiate bit range and add it
            length += (byte & ((1 << 7) - 1)) << bit_range

            # Check MSB and if it is zero, this is the last length byte and we are ready to read string
            if byte & (1 << 7) == 0:
                break
        else:
            # For-else statements in Python are useful! This will be only happen if the for successfully completes
            raise NRBFError('NRBF LengthPrefixedString overflow')

        # Read the string
        return self.stream.read(length).decode('utf-8')

    # endregion

    # region Primitive Array reader functions

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Boolean)
    def _readBoolArray(self, length):
        return struct.unpack('<{0}?'.format(length), self.stream.read(length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Byte)
    def _readByteArray(self, length):
        return struct.unpack('<{0}B'.format(length), self.stream.read(length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Double)
    def _readDoubleArray(self, length):
        return struct.unpack('<{0}d'.format(length), self.stream.read(8 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Int16)
    def _readInt16Array(self, length):
        return struct.unpack('<{0}h'.format(length), self.stream.read(2 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Int32)
    def _readInt32Array(self, length):
        return struct.unpack('<{0}i'.format(length), self.stream.read(4 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Int64)
    def _readInt64Array(self, length):
        return struct.unpack('<{0}q'.format(length), self.stream.read(8 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.SByte)
    def _readSByteArray(self, length):
        return struct.unpack('<{0}b'.format(length), self.stream.read(length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.Single)
    def _readSingleArray(self, length):
        return struct.unpack('<{0}f'.format(length), self.stream.read(4 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.UInt16)
    def _readUInt16Array(self, length):
        return struct.unpack('<{0}H'.format(length), self.stream.read(2 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.UInt32)
    def _readUInt32Array(self, length):
        return struct.unpack('<{0}I'.format(length), self.stream.read(4 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.UInt64)
    def _readUInt64Array(self, length):
        return struct.unpack('<{0}Q'.format(length), self.stream.read(8 * length))

    @_registerReader(_PrimitiveTypeArrayReaders, PrimitiveType.String)
    def _readStringArray(self, length):
        return [self._readString() for i in range(length)]

    # endregion

    # region AdditionalInfo reader functions

    # Note: Remaining reader functions are attached in other sections to existing functions

    # The ClassTypeInfo structure is read and ignored
    @_registerReader(_AdditionalInfoReaders, BinaryType.Class)
    def _readClassTypeInfo(self):
        name = self._readString()
        libraryID = self._readInt32()

        return (name, libraryID)

    # endregion

    # region Record reader functions

    def _readRecord(self):
        recordType = self._readByte()

        return self._RecordTypeReaders[recordType](self)

    def _readPrimitive(self, primitiveType):
        return self._PrimitiveTypeReaders[primitiveType](self)

    @_registerReader(_RecordTypeReaders, RecordType.SerializedStreamHeader)
    def _readSerializationHeaderRecord(self):
        self.rootID = self._readInt32()
        self.headerID = self._readInt32()
        majorVersion, minorVersion = self._readInt32(), self._readInt32()

        if majorVersion != 1 or minorVersion != 0:
            raise NRBFError('Major and minor version for serialization header is incorrect: {0} {1}'.format(
                majorVersion, minorVersion))

        if self.rootID == 0:
            raise NotImplementedError(
                'Root ID is zero indicating that a BinaryMethodCall is available. Not implemented yet')

    @_registerReader(_RecordTypeReaders, RecordType.ClassWithId)
    def _readClassWithId(self):
        objectID = self._readInt32()
        metadataID = self._readInt32()
        cls = self.classByID[metadataID]

        # Only instance where the objectID does NOT equal the class ID is here!
        return self._readClassMembers(cls(), objectID)

    @_registerReader(_RecordTypeReaders, RecordType.SystemClassWithMembers)
    def _readSystemClassWithMembers(self):
        cls = self._readClassInfo(isSystemClass=True)
        cls._typeInfo = None

        return self._readClassMembers(cls(), cls._id)

    @_registerReader(_RecordTypeReaders, RecordType.ClassWithMembers)
    def _readClassWithMembers(self):
        cls = self._readClassInfo(isSystemClass=False)
        cls._typeInfo = None
        libraryID = self._readInt32()

        return self._readClassMembers(cls(), cls._id, libraryID)

    @_registerReader(_RecordTypeReaders, RecordType.SystemClassWithMembersAndTypes)
    def _readSystemClassWithMembersAndTypes(self):
        cls = self._readClassInfo(isSystemClass=True)
        self._readMemberTypeInfo(cls)

        return self._readClassMembers(cls(), cls._id)

    @_registerReader(_RecordTypeReaders, RecordType.ClassWithMembersAndTypes)
    def _readClassWithMembersAndTypes(self):
        cls = self._readClassInfo(isSystemClass=False)
        self._readMemberTypeInfo(cls)
        libraryID = self._readInt32()

        return self._readClassMembers(cls(), cls._id, libraryID)

    @_registerReader(_RecordTypeReaders, RecordType.BinaryObjectString)
    def _readBinaryObjectString(self):
        objectID = self._readInt32()
        string = self._readString()
        self.objectsByID[objectID] = string

        return string

    @_registerReader(_RecordTypeReaders, RecordType.MemberPrimitiveTyped)
    def _readMemberPrimitiveTyped(self):
        primitiveType = self._readInt32()
        value = self._PrimitiveTypeReaders[primitiveType](self)

        return value

    @_registerReader(_RecordTypeReaders, RecordType.BinaryArray)
    def _readBinaryArray(self):
        objectID = self._readInt32()
        arrayType = self._readByte()
        rank = self._readInt32()
        lengths = [self._readInt32() for i in range(rank)]

        # The lower bounds are ignored currently
        # Not sure of the implications or purpose of this
        if arrayType in [BinaryArrayType.SingleOffset, BinaryArrayType.JaggedOffset, BinaryArrayType.RectangularOffset]:
            lowerBounds = [self._readInt32() for i in range(rank)]

        binaryType = self._readByte()
        additionalInfo = self._AdditionalInfoReaders[binaryType](self)

        # Get total length of items that we need to read
        # This is just the product of all the elements in the lengths array
        length = reduce(lambda x, y: x * y, lengths)

        # If the items are primitives, use primitive array readers
        # Otherwise, the items will be objects and should be read by reading records
        if binaryType == BinaryType.Primitive:
            array = self._PrimitiveTypeArrayReaders[additionalInfo](self, length)
        else:
            array = self._readObjectArray(length, objectID)

        # For a multidimensional array, take the 1D array that was read and convert it to ND
        if arrayType in [BinaryArrayType.Rectangular, BinaryArrayType.RectangularOffset]:
            array = convert1DArrayND(array, lengths)

        # Save the object by ID
        # Only required for primitive because _readObjectArray saves the ID for you
        # But we just overwrite it regardless
        self.objectsByID[objectID] = array

        return array

    # When object's with an object ID are encountered above, they are added to the _objectsByID dictionary.
    # A MemberReference object contains the object ID that the reference refers to. These references are
    # resolved at the end.
    @_registerReader(_RecordTypeReaders, RecordType.MemberReference)
    def _readMemberReference(self):
        # objectID
        ref = Reference(self._readInt32())

        return ref

    @_registerReader(_RecordTypeReaders, RecordType.MessageEnd)
    def _readMessageEnd(self):
        return MessageEnd()

    @_registerReader(_RecordTypeReaders, RecordType.BinaryLibrary)
    def _readBinaryLibrary(self):
        libraryID = self._readInt32()
        libraryName = self._readString()
        library = BinaryLibrary(libraryID, libraryName, {})
        self.binaryLibraries[libraryID] = library

        return library

    @_registerReader(_RecordTypeReaders, RecordType.ObjectNullMultiple256)
    def _readObjectNullMultiple256(self):
        # Count
        return ObjectNullMultiple(self._readByte())

    @_registerReader(_RecordTypeReaders, RecordType.ObjectNullMultiple)
    def _readObjectNullMultiple(self):
        # Count
        return ObjectNullMultiple(self._readInt32())

    @_registerReader(_RecordTypeReaders, RecordType.ArraySinglePrimitive)
    def _readArraySinglePrimitive(self):
        objectID, length = self._readArrayInfo()
        primitiveType = self._readByte()

        array = self._PrimitiveTypeArrayReaders[primitiveType](self, length)
        self.objectsByID[objectID] = array

        return array

    @_registerReader(_RecordTypeReaders, RecordType.ArraySingleObject)
    def _read_ArraySingleObject(self):
        objectID, length = self._readArrayInfo()

        array = self._readObjectArray(length, objectID)

        return array

    @_registerReader(_RecordTypeReaders, RecordType.ArraySingleString)
    def _readArraySingleString(self):
        objectID, length = self._readArrayInfo()
        primitiveType = PrimitiveType.String

        array = self._PrimitiveTypeArrayReaders[primitiveType](self, length)
        self.objectsByID[objectID] = array

        return array

    @_registerReader(_RecordTypeReaders, RecordType.MethodCall)
    def _read_MethodCall(self):
        raise NotImplementedError('MethodCall')

    @_registerReader(_RecordTypeReaders, RecordType.MethodReturn)
    def _read_MethodReturn(self):
        raise NotImplementedError('MethodReturn')

    # endregion

    # region Read helper classes

    def _readArrayInfo(self):
        # objectID and array length
        return (self._readInt32(), self._readInt32())

    # Reads a ClassInfo structure and creates and saves a new namedlist object with the same member and ClassInfo
    # specifics
    def _readClassInfo(self, isSystemClass=False):
        objectID = self._readInt32()
        className = self._readString()
        memberCount = self._readInt32()
        memberNames = [sanitizeIdentifier(self._readString()) for i in range(memberCount)]

        # Create namedlist that will act as the class
        # Set object ID for the class and save it by the object ID for later
        cls = namedlist(sanitizeIdentifier(className), memberNames, default=None)

        # Class ID is a special identifier to represent the class itself but not the data within of it
        # For instance, you can declare the class once but then instantiate it multiple times
        # The object ID is the unique identifier for the object itself and cannot be repeated
        # In this instance, the classID and objectID are the same for the first instantiation
        # The only way to get these different is to have a ClassWithID record and then this is
        # set manually
        cls._classID = cls._id = objectID
        cls._isSystemClass = isSystemClass
        self.classByID[objectID] = cls

        return cls

    def _readMemberTypeInfo(self, cls):
        binaryTypes = [self._readByte() for member in cls._fields]
        additionalInfo = [self._AdditionalInfoReaders[type](self) for type in binaryTypes]

        # Combine the binary types and additional info into one tuple
        # This gets saved to the class object because it will be used to set the members
        # Also, if there is a ClassWithId object then it won't have type information but only
        # the class ID so we will need to retain this information.
        cls._typeInfo = tuple(zip(binaryTypes, additionalInfo))

    # Reads members or array elements into the 'obj' pre-allocated list or class instance
    def _readClassMembers(self, obj, objectID, libraryID=None):
        index = 0
        while index < len(obj._fields):
            # If typeinfo is not defined, as is the case for ClassWithMembers and SystemClassWithMembers,
            # then assume it is an object that can be read
            # Not sure if this is a safe assumption because Microsoft isn't entirely clear when the member type
            # information is 'unnecessary'
            if obj._typeInfo is None:
                binaryType, additionalInfo = BinaryType.Object, None
            else:
                binaryType, additionalInfo = obj._typeInfo[index]

            if binaryType == BinaryType.Primitive:
                value = self._readPrimitive(additionalInfo)
            else:
                value = self._readRecord()

                if isinstance(value, BinaryLibrary):
                    # BinaryLibrary can precede the actual member
                    # Continue on to the actual member
                    continue
                elif isinstance(value, ObjectNullMultiple):
                    # Skip a given number of indices
                    index += value.count
                    continue
                elif isinstance(value, Reference):
                    value.parent = obj
                    value.indexInParent = index

            obj[index] = value
            index += 1

        # If this object is a .NET collection (e.g. a Generic dict or list) which can be
        # replaced by a native Python type, insert a collection _Reference instead of the raw
        # object which will be resolved later in read() using a "collection resolver"
        if getattr(obj.__class__, '_isSystemClass', False):
            for name, resolver in self._collectionResolvers:
                if obj.__class__.__name__.startswith('System_Collections_Generic_%s_' % name):
                    obj = Reference(objectID, collectionResolver=resolver, originalObj=obj)
                    self._collectionReferences.append(obj)
                    break

        self.objectsByID[objectID] = obj

        # Use libraryID to append the class to that binary library
        # This is particularly useful when saving the items again so that you save all of a binary library at once
        if libraryID:
            self.binaryLibraries[libraryID].objects[objectID] = obj

        return obj

    def _readObjectArray(self, length, objectID):
        array = [None] * length

        index = 0
        while index < length:
            value = self._readRecord()

            if isinstance(value, BinaryLibrary):
                # BinaryLibrary can precede the actual member
                # Continue on to the actual member
                continue
            elif isinstance(value, ObjectNullMultiple):
                # Skip a given number of indices
                index += value.count
                continue
            elif isinstance(value, Reference):
                value.parent = array
                value.indexInParent = index

            # Save the object by ID
            self.objectsByID[objectID] = array

            array[index] = value
            index += 1

        return array

    # endregion

    # region Resolve references functions

    # Convert a _Reference representing a MemberReference into its referenced object
    def _resolveSimpleReference(self, reference):
        if reference.resolved:
            return

        replacement = self.objectsByID[reference._id]
        reference.parent[reference.indexInParent] = replacement

        # Not sure when this would happen, doesn't hurt though!
        reference.resolved = True

    # Convert a _Reference representing a .NET dictionary collection into a Python dict
    def _resolveDictReference(self, reference):
        originalObj = reference.originalObj

        # If the key-value pairs of the dict are itself a Reference, then resolve those first
        if isinstance(originalObj.KeyValuePairs, Reference):
            self._resolveSimpleReference(originalObj.KeyValuePairs)

        replacement = {}

        for item in originalObj.KeyValuePairs:
            try:
                # If any key is a _Reference, it must be resolved first
                # (value _References will be resolved later)
                if isinstance(item.key, Reference):
                    self._resolveSimpleReference(item.key)

                assert item.key not in replacement
                replacement[item.key] = item.value
            except (AssertionError, TypeError):
                # Not all .NET dictionaries can be converted to Python dicts
                # If the conversion fails, just proceed w/ the original object
                replacement = originalObj
                break
        else:
            # Assuming that the for loop completed successfully, indicating that all
            # of the dictionary objects were converted,
            # Then we need to fix the dictionary values for References
            for key, value in replacement.items():
                if isinstance(value, Reference):
                    value.parent = replacement
                    value.index_in_parent = key

        return replacement

    # Convert a Reference representing a .NET list collection into a Python list
    def _resolveListReference(self, reference):
        originalObj = reference.originalObj

        # If the components of the list are itself a Reference, then resolve those first
        if isinstance(originalObj.items, Reference):
            self._resolveSimpleReference(originalObj.items)

        replacement = originalObj.items

        # Update parent for all replacement elements if they are references
        for element in replacement:
            if isinstance(element, Reference):
                element.parent = replacement

        return replacement

    _collectionResolvers = (
        ('Dictionary', _resolveDictReference),
        ('List', _resolveListReference)
    )

    # endregion



from pypdn.reader import read, BlendType, Layer, LayeredImage
from pypdn.util import JSONEncoder
from pypdn.nrbf import NRBF

__all__ = ['read', 'JSONEncoder', 'NRBF', 'BlendType', 'Layer', 'LayeredImage']



from setuptools import setup, find_packages
import os

currentPath = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(currentPath, 'README.rst'), 'r') as f:
    long_description = f.read()

long_description = '\n' + long_description
setup(name='pypdn',
      version='1.0.1',
      description='Python package to read and write Paint.NET (PDN) images.',
      long_description=long_description,
      author='Addison Elliott',
      author_email='addison.elliott@gmail.com',
      url='https://github.com/addisonElliott/pypdn',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Topic :: Scientific/Engineering',
          'License :: OSI Approved :: MIT License',
          "Programming Language :: Python",
          'Programming Language :: Python :: 3',
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Operating System :: OS Independent"

      ],
      keywords='paint dot net paintdotnet dotnet .net pdn library read write save layered images image editor',
      project_urls={
          'Documentation': 'https://github.com/addisonElliott/pypdn',
          'Source': 'https://github.com/addisonElliott/pypdn',
          'Tracker': 'https://github.com/addisonElliott/pypdn/issues',
      },
      python_requires='>=3',
      packages=find_packages(),
      package_data={
        'tests': ['data/*']
      },
      license='MIT License',
      install_requires=[
          'numpy', 'scikit-image', 'aenum']
      )



from datetime import datetime, timedelta
import json
import re
from collections import OrderedDict
from decimal import Decimal
from aenum import IntEnum
from keyword import iskeyword
import pypdn.nrbf

from pypdn.namedlist import namedlist, namedtuple


class RecordType(IntEnum):
    SerializedStreamHeader = 0
    ClassWithId = 1
    SystemClassWithMembers = 2
    ClassWithMembers = 3
    SystemClassWithMembersAndTypes = 4
    ClassWithMembersAndTypes = 5
    BinaryObjectString = 6
    BinaryArray = 7
    MemberPrimitiveTyped = 8
    MemberReference = 9
    ObjectNull = 10
    MessageEnd = 11
    BinaryLibrary = 12
    ObjectNullMultiple256 = 13
    ObjectNullMultiple = 14
    ArraySinglePrimitive = 15
    ArraySingleObject = 16
    ArraySingleString = 17
    MethodCall = 21
    MethodReturn = 22


class PrimitiveType(IntEnum):
    Boolean = 1
    Byte = 2
    Char = 3
    Decimal = 5
    Double = 6
    Int16 = 7
    Int32 = 8
    Int64 = 9
    SByte = 10
    Single = 11
    TimeSpan = 12
    DateTime = 13
    UInt16 = 14
    UInt32 = 15
    UInt64 = 16
    Null = 17
    String = 18


class BinaryType(IntEnum):
    Primitive = 0
    String = 1
    Object = 2
    SystemClass = 3
    Class = 4
    ObjectArray = 5
    StringArray = 6
    PrimitiveArray = 7


class BinaryArrayType(IntEnum):
    Single = 0
    Jagged = 1
    Rectangular = 2
    SingleOffset = 3
    JaggedOffset = 4
    RectangularOffset = 5


# Given an identifier string, sanitize the string such that it is suitable to pass to namedlist
def sanitizeIdentifier(identifier):
    # Replace anything that is not an alphanumeric character or underscore with an underscore
    # Also, remove any leading numbers because you cannot reference an member starting with a number
    identifier = re.sub(r'[^a-z0-9_]', '_', identifier, flags=re.IGNORECASE).lstrip('0123456789_')

    # Append an underscore if the identifier is a keyword
    if iskeyword(identifier):
        identifier += '_'

    return identifier


# Take a 1D array and convert it to a N-dimensional array with dimensions given by the arguments dims
# This function uses recursion to accomplish the task so index must be a mutable list with one number inside of it
# The element in the list will be the current index and will be incremented in the recursion
def convert1DArrayND(array1d, dims, index=[0]):
    if len(dims) == 1:
        array = list(array1d[index[0]:index[0] + dims[0]])
        index[0] += dims[0]
        return array
    else:
        return [convert1DArrayND(array1d, dims[1:], index) for x in range(dims[0])]


BinaryLibrary = namedlist('BinaryLibrary', ['_id', 'name', 'objects'], default=None)
Reference = namedlist('Reference', ['_id', 'parent', 'indexInParent', 'resolved', 'collectionResolver', 'originalObj'],
                      default=None)
MessageEnd = namedlist('MessageEnd', [])
ObjectNullMultiple = namedtuple('ObjectNullMultiple', 'count')


# Custom JSONEncoder to convert NRBF class or any of the subclasses into JSON
# This class DOES handle circular references, something that is common in the .NET world
class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, pypdn.nrbf.NRBF):
            d = OrderedDict(SerializationHeader={'rootID': o.rootID, 'headerID': o.headerID,
                                                 'majorVersion': 1, 'minorVersion': 0})

            # Attach root
            d['Root'] = o.getRoot()

            # Add all binary libraries
            d['BinaryLibraries'] = o.binaryLibraries

            # Attach classes and objects by ID
            d['Objects'] = o.objectsByID

            return d
        elif isinstance(o, Reference):
            # We have to handle the reference specially because there is the parent field that will
            # cause circular dependencies
            return OrderedDict(_class_name=o.__class__.__name__, id=o._id)
        elif hasattr(o, '_asdict'):
            d = OrderedDict(_class_name=o.__class__.__name__)  # prepend the class name
            if hasattr(o, '_id'):
                d['_id'] = o._id
            d.update(o._asdict())
            return d
        elif isinstance(o, (datetime, timedelta)):
            return str(o)
        elif isinstance(o, Decimal):
            return repr(o)

        return super().default(o)

    def afterItem(self, o):
        # pass
        if hasattr(o, '_asdict'):
            o._ref_count = 0
